"""Core network client for the Eportal login flow."""

import gzip
import io
import json
import logging
import re
import time
from collections.abc import Callable
from dataclasses import dataclass
from urllib.parse import parse_qs, quote, urljoin, urlparse

import requests

from .config import ResolvedConfig
from .logging_utils import mask_value, sanitize_url, truncate_text
from .security import encryptPassword

PreparedData = dict[str, str]
SleepFunc = Callable[[float], None]


@dataclass(slots=True)
class PreparedRequest:
    """Request information assembled before sending."""

    url: str
    headers: dict[str, str]
    data: PreparedData


class CampusLoginClient:
    """Campus network login client."""

    def __init__(
        self,
        config: ResolvedConfig,
        logger: logging.Logger,
        *,
        session_factory: Callable[[], requests.Session] = requests.Session,
        request_get: Callable[..., requests.Response] = requests.get,
        sleep_func: SleepFunc = time.sleep,
    ) -> None:
        self.config = config
        self.logger = logger
        self.session: requests.Session | None = None
        self._session_factory = session_factory
        self._request_get = request_get
        self._sleep = sleep_func
        self._last_user_index: str | None = None

    def _create_session(self) -> requests.Session:
        session = self._session_factory()
        session.headers.update(
            {
                "User-Agent": (
                    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) "
                    "Chrome/135.0.0.0 Safari/537.36"
                ),
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
                "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8",
                "Accept-Encoding": "gzip, deflate",
            }
        )
        return session

    def _decode_response(self, response: requests.Response) -> str:
        """Decode response bodies that may be compressed or use GBK encodings."""
        content_encoding = response.headers.get("Content-Encoding", "").lower()
        payload = response.content

        if content_encoding == "gzip":
            try:
                with gzip.GzipFile(fileobj=io.BytesIO(response.content)) as handle:
                    payload = handle.read()
            except OSError as exc:
                self.logger.warning("gzip 解压失败，继续尝试按原始内容解析: %s", exc)

        for encoding in ("utf-8", "gb18030", "gbk"):
            try:
                return payload.decode(encoding)
            except UnicodeDecodeError:
                continue

        return payload.decode("utf-8", errors="replace")

    @staticmethod
    def _extract_redirect_url(content: str) -> str | None:
        patterns = (
            r"top\.self\.location\.href='(.*?)'",
            r'top\.self\.location\.href="(.*?)"',
            r"location\.href\s*=\s*'(.*?)'",
            r'location\.href\s*=\s*"(.*?)"',
        )
        for pattern in patterns:
            match = re.search(pattern, content, re.IGNORECASE)
            if match:
                return match.group(1)
        return None

    @staticmethod
    def _parse_login_response(content: str) -> tuple[bool, str]:
        try:
            payload = json.loads(content)
        except json.JSONDecodeError:
            lowered = content.lower()
            if "success" in lowered or "成功" in content:
                return True, "登录成功（通过响应文本识别）"
            detail = truncate_text(content, 80)
            return False, f"无法识别登录响应，请使用 --verbose 查看详情。响应片段: {detail}"

        if payload.get("result") == "success":
            message = payload.get("message") or "登录成功"
            return True, message

        message = payload.get("message") or "登录失败"
        return False, message

    def check_internet(self, timeout: int = 5) -> bool:
        """Check whether the configured probe URL is reachable."""
        try:
            response = self._request_get(
                self.config.check_url,
                timeout=timeout,
                allow_redirects=True,
            )
        except requests.exceptions.Timeout:
            self.logger.debug("网络检测超时: %s", self.config.check_url)
            return False
        except requests.exceptions.RequestException as exc:
            self.logger.debug("网络检测失败: %s", exc)
            return False

        if 200 <= response.status_code < 300 and response.content:
            self.logger.debug("网络检测成功: %s", self.config.check_url)
            return True

        self.logger.debug("网络检测返回异常状态: %s", response.status_code)
        return False

    def probe_auth_trigger(self, timeout: int = 10) -> tuple[bool, str]:
        """Probe the trigger endpoint and report whether a login page is discoverable."""
        session = self._create_session()
        try:
            response = session.get(self.config.trigger_url, timeout=timeout, allow_redirects=True)
            login_url = self._find_login_page_url(response)
            if login_url:
                return True, sanitize_url(login_url)
            user_index, status = self._discover_user_index_from_response(response)
            if status == "online" and user_index:
                return True, "当前已在线，并发现有效会话。"
            return False, f"认证触发地址返回了 HTTP {response.status_code}，但未找到登录页跳转。"
        except requests.exceptions.RequestException as exc:
            return False, f"无法访问认证触发地址: {exc}"
        finally:
            session.close()

    def build_page_info_request(self, login_page_url: str, query_string: str) -> PreparedRequest:
        parsed = urlparse(login_page_url)
        origin = f"{parsed.scheme}://{parsed.netloc}"
        return PreparedRequest(
            url=urljoin(login_page_url, "/eportal/InterFace.do?method=pageInfo"),
            headers={
                "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
                "Referer": login_page_url,
                "Origin": origin,
                "Accept": "*/*",
            },
            data={"queryString": quote(query_string)},
        )

    def build_login_request(
        self,
        login_page_url: str,
        query_string: str,
        encrypted_password: str,
        service: str = "",
    ) -> PreparedRequest:
        parsed = urlparse(login_page_url)
        origin = f"{parsed.scheme}://{parsed.netloc}"
        return PreparedRequest(
            url=urljoin(login_page_url, "/eportal/InterFace.do?method=login"),
            headers={
                "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
                "Referer": login_page_url,
                "Origin": origin,
            },
            data={
                "userId": self.config.username,
                "password": encrypted_password,
                "service": service,
                "queryString": quote(query_string),
                "operatorPwd": "",
                "operatorUserId": "",
                "validcode": "",
                "passwordEncrypt": "true",
            },
        )

    def build_interface_request(self, method: str, data: PreparedData) -> PreparedRequest:
        """Build a portal interface request that targets a method under InterFace.do."""
        parsed = urlparse(self.config.trigger_url)
        origin = f"{parsed.scheme}://{parsed.netloc}"
        endpoint = urljoin(self.config.trigger_url, f"/eportal/InterFace.do?method={method}")
        return PreparedRequest(
            url=endpoint,
            headers={
                "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8",
                "Origin": origin,
                "Referer": self.config.trigger_url,
                "Accept": "*/*",
            },
            data=data,
        )

    def _parse_json_response(
        self,
        response: requests.Response,
        *,
        action: str,
    ) -> dict | None:
        content = self._decode_response(response)
        try:
            payload = json.loads(content)
        except json.JSONDecodeError as exc:
            self.logger.error("解析%s响应失败: %s", action, exc)
            self.logger.debug("%s响应片段: %s", action, truncate_text(content))
            return None

        if isinstance(payload, dict):
            return payload

        self.logger.error("%s响应不是 JSON 对象。", action)
        self.logger.debug("%s响应内容: %s", action, truncate_text(str(payload)))
        return None

    def _post_interface_json(
        self,
        method: str,
        data: PreparedData,
        *,
        timeout: int = 15,
    ) -> dict | None:
        if self.session is None:
            raise RuntimeError("portal 请求前尚未初始化 session。")

        request = self.build_interface_request(method, data)
        try:
            response = self.session.post(
                request.url,
                headers=request.headers,
                data=request.data,
                timeout=timeout,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            self.logger.error("请求 %s 失败: %s", method, exc)
            return None

        payload = self._parse_json_response(response, action=method)
        if payload is not None:
            self.logger.debug("%s 响应字段: %s", method, ", ".join(sorted(payload.keys())))
        return payload

    @staticmethod
    def _looks_like_not_online(message: str) -> bool:
        lowered = message.lower()
        patterns = ("未在线", "不在线", "已下线", "不存在", "not online", "offline")
        return any(pattern in lowered or pattern in message for pattern in patterns)

    @staticmethod
    def _looks_like_incomplete_user_info(message: str) -> bool:
        patterns = ("用户信息不完整", "信息不完整", "please retry")
        lowered = message.lower()
        return any(pattern in message or pattern in lowered for pattern in patterns)

    @staticmethod
    def _extract_user_index_from_url(url: str) -> str | None:
        parsed = urlparse(url)
        return parse_qs(parsed.query).get("userIndex", [None])[0]

    @staticmethod
    def _extract_user_index_from_content(content: str) -> str | None:
        patterns = (
            r"userIndex=([0-9a-fA-F]+)",
            r'"userIndex"\s*:\s*"([0-9a-fA-F]+)"',
            r"userIndex\s*[:=]\s*['\"]?([0-9a-fA-F]+)",
        )
        for pattern in patterns:
            match = re.search(pattern, content)
            if match:
                return match.group(1)
        return None

    @staticmethod
    def _looks_like_login_page_url(url: str) -> bool:
        parsed = urlparse(url)
        if not parsed.path.endswith("/eportal/index.jsp"):
            return False
        query = parse_qs(parsed.query)
        return any(key in query for key in ("wlanuserip", "nasip", "mac"))

    def _iter_response_urls_and_content(
        self,
        response: requests.Response,
    ) -> tuple[list[str], list[str]]:
        urls: list[str] = []
        contents: list[str] = []
        for item in [*response.history, response]:
            if item.url:
                urls.append(item.url)

            location = item.headers.get("Location")
            if location:
                urls.append(urljoin(item.url or self.config.trigger_url, location))

            content = self._decode_response(item)
            contents.append(content)

            redirect_url = self._extract_redirect_url(content)
            if redirect_url:
                urls.append(urljoin(item.url or self.config.trigger_url, redirect_url))

        return urls, contents

    def _find_login_page_url(self, response: requests.Response) -> str | None:
        urls, _ = self._iter_response_urls_and_content(response)
        for url in urls:
            if self._looks_like_login_page_url(url):
                return url
        return None

    def _discover_user_index_from_response(
        self,
        response: requests.Response,
    ) -> tuple[str | None, str]:
        urls, contents = self._iter_response_urls_and_content(response)

        for url in urls:
            user_index = self._extract_user_index_from_url(url)
            if user_index:
                return user_index, "online"

        for content in contents:
            user_index = self._extract_user_index_from_content(content)
            if user_index:
                return user_index, "online"

        for url in urls:
            if self._looks_like_login_page_url(url):
                return None, "not_online"

        return None, "error"

    def _discover_user_index(self) -> tuple[str | None, str]:
        self.logger.info("正在发现当前在线会话的 userIndex")
        try:
            response = self.session.get(self.config.trigger_url, timeout=15, allow_redirects=True)
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            self.logger.error("访问认证入口失败: %s", exc)
            return None, "error"

        user_index, status = self._discover_user_index_from_response(response)
        if status == "online" and user_index:
            self.logger.info("已从认证入口发现 userIndex")
            return user_index, status
        if status == "not_online":
            self.logger.info("认证入口当前返回登录页，说明当前未在线。")
            return None, status

        self.logger.error("无法从认证入口响应中提取 userIndex。")
        urls, _ = self._iter_response_urls_and_content(response)
        if urls:
            self.logger.debug("认证入口候选 URL: %s", " | ".join(sanitize_url(url) for url in urls))
        return None, status

    def _get_login_page_url(self) -> str | None:
        self.logger.info("正在访问认证触发地址")
        self.logger.debug("认证触发地址: %s", self.config.trigger_url)

        try:
            response = self.session.get(self.config.trigger_url, timeout=15, allow_redirects=True)
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            self.logger.error("访问认证触发地址失败: %s", exc)
            return None

        login_url = self._find_login_page_url(response)
        if login_url:
            self.logger.info("已发现登录页面")
            self.logger.debug("登录页面 URL: %s", sanitize_url(login_url))
            return login_url

        self.logger.error("无法从认证触发响应中提取登录页面 URL。")
        self.logger.debug("触发入口最终 URL: %s", sanitize_url(response.url))
        return None

    def _get_page_info(self, login_page_url: str, query_string: str) -> dict | None:
        request = self.build_page_info_request(login_page_url, query_string)
        self.logger.info("正在获取 pageInfo 公钥信息")
        try:
            response = self.session.post(
                request.url,
                headers=request.headers,
                data=request.data,
                timeout=15,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            self.logger.error("请求 pageInfo 失败: %s", exc)
            return None

        content = self._decode_response(response)
        try:
            payload = json.loads(content)
        except json.JSONDecodeError as exc:
            self.logger.error("解析 pageInfo 响应失败: %s", exc)
            self.logger.debug("pageInfo 响应片段: %s", truncate_text(content))
            return None

        self.logger.debug("pageInfo 字段: %s", ", ".join(sorted(payload.keys())))
        return payload

    def _submit_login(
        self,
        login_page_url: str,
        query_string: str,
        encrypted_password: str,
        service: str = "",
    ) -> tuple[bool, str]:
        request = self.build_login_request(
            login_page_url,
            query_string,
            encrypted_password,
            service,
        )
        self.logger.info("正在提交登录请求")

        try:
            response = self.session.post(
                request.url,
                headers=request.headers,
                data=request.data,
                timeout=20,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as exc:
            return False, f"登录请求失败: {exc}"

        content = self._decode_response(response)
        self.logger.debug("登录响应片段: %s", truncate_text(content))
        success, message = self._parse_login_response(content)
        if success:
            payload = self._parse_json_response(response, action="login")
            if payload is not None:
                self._last_user_index = payload.get("userIndex") or self._last_user_index
        return success, message

    def get_online_user_info(self, user_index: str) -> dict | None:
        """Fetch online session information for a userIndex."""
        return self._post_interface_json(
            "getOnlineUserInfo",
            {"userIndex": user_index},
        )

    def _get_online_user_info_with_retry(
        self,
        initial_user_index: str,
    ) -> tuple[dict | None, str]:
        user_index = initial_user_index
        for attempt in range(2):
            info = self.get_online_user_info(user_index)
            if info is None:
                return None, user_index

            if info.get("result") == "success":
                return info, user_index

            message = str(info.get("message") or "获取在线状态失败")
            if not self._looks_like_incomplete_user_info(message) or attempt == 1:
                return info, user_index

            self.logger.info("在线状态信息暂不完整，等待后重试一次。")
            self._sleep(1)
            rediscovered_user_index, discovery_status = self._discover_user_index()
            if discovery_status == "not_online":
                return {"result": "fail", "message": "当前用户不在线"}, user_index
            if rediscovered_user_index:
                user_index = rediscovered_user_index

        return None, user_index

    def logout(self, user_index: str | None = None) -> bool:
        """Log out the current campus network session if one is active."""
        self.logger.info("开始校园网退出登录")
        self.session = self._create_session()
        try:
            resolved_user_index = user_index or self._last_user_index
            discovery_status = "online"
            if not resolved_user_index:
                resolved_user_index, discovery_status = self._discover_user_index()

            if discovery_status == "not_online":
                self.logger.info("当前未检测到在线会话，无需退出。")
                return True

            if not resolved_user_index:
                self.logger.error(
                    "无法确定 userIndex。"
                    "请在已在线状态下运行，或手动提供 --user-index。"
                )
                return False

            info, resolved_user_index = self._get_online_user_info_with_retry(resolved_user_index)
            if info is None:
                return False

            if info.get("result") != "success":
                message = str(info.get("message") or "获取在线状态失败")
                if self._looks_like_not_online(message):
                    self.logger.info("当前未检测到在线会话，无需退出。")
                    return True
                self.logger.error("获取在线状态失败: %s", message)
                return False

            self.logger.info("检测到在线会话，正在提交退出请求。")
            payload = self._post_interface_json(
                "logout",
                {"userIndex": resolved_user_index},
            )
            if payload is None:
                return False

            if payload.get("result") == "success":
                message = str(payload.get("message") or "下线成功")
                self.logger.info(message)
                self._last_user_index = None
                return True

            message = str(payload.get("message") or "退出登录失败")
            if self._looks_like_not_online(message):
                self.logger.info("当前未检测到在线会话，无需退出。")
                return True

            self.logger.error("退出登录失败: %s", message)
            return False
        except Exception as exc:  # pragma: no cover - safety net
            self.logger.exception("退出登录过程发生未处理异常: %s", exc)
            return False
        finally:
            if self.session is not None:
                self.session.close()
                self.session = None

    def login(self) -> bool:
        """Execute the full login flow."""
        self.logger.info("开始校园网登录")
        self.logger.debug("用户名: %s", mask_value(self.config.username))

        self.session = self._create_session()
        try:
            login_page_url = self._get_login_page_url()
            if not login_page_url:
                return False

            parsed = urlparse(login_page_url)
            query_params = parse_qs(parsed.query)
            query_string = parsed.query
            mac_address = query_params.get("mac", [None])[0]
            service = query_params.get("service", [""])[0]

            if not mac_address:
                self.logger.error("登录页参数缺少 mac，无法继续加密密码。")
                return False

            self.logger.debug("已解析登录页查询串，mac: %s", mask_value(mac_address))

            try:
                self.session.get(login_page_url, timeout=15)
            except requests.exceptions.RequestException as exc:
                self.logger.warning("访问登录页失败，但继续尝试登录: %s", exc)

            page_info = self._get_page_info(login_page_url, query_string)
            if not page_info:
                return False

            public_key_modulus = page_info.get("publicKeyModulus")
            public_key_exponent = page_info.get("publicKeyExponent", "10001")

            if not public_key_modulus:
                self.logger.error("pageInfo 缺少 publicKeyModulus。")
                return False

            self.logger.info("正在加密密码")
            try:
                encrypted_password = encryptPassword(
                    self.config.password,
                    public_key_exponent,
                    public_key_modulus,
                    mac_address,
                )
            except Exception as exc:  # pragma: no cover - delegated RSA logic
                self.logger.error("密码加密失败: %s", exc)
                return False

            if not encrypted_password:
                self.logger.error("密码加密结果为空。")
                return False

            success, message = self._submit_login(
                login_page_url,
                query_string,
                encrypted_password,
                service,
            )
            if success:
                self.logger.info("登录成功")
                if message and message != "登录成功":
                    self.logger.debug("登录消息: %s", message)
                return True

            self.logger.error("登录失败: %s", message)
            return False
        except Exception as exc:  # pragma: no cover - safety net
            self.logger.exception("登录过程发生未处理异常: %s", exc)
            return False
        finally:
            if self.session is not None:
                self.session.close()
                self.session = None

    def login_with_retry(self) -> bool:
        """Retry login with exponential backoff."""
        for attempt in range(1, self.config.max_retries + 1):
            self.logger.info("登录尝试 %s/%s", attempt, self.config.max_retries)
            if self.login():
                self._sleep(2)
                if self.check_internet():
                    self.logger.info("登录后网络验证成功")
                    return True

                self.logger.warning("登录成功但网络验证尚未通过，稍后再试一次。")
                self._sleep(3)
                if self.check_internet():
                    self.logger.info("延迟复检成功")
                    return True

            if attempt < self.config.max_retries:
                wait_time = 2 ** attempt
                self.logger.info("等待 %s 秒后重试", wait_time)
                self._sleep(wait_time)

        self.logger.error(
            "登录失败，已重试 %s 次。建议使用 `campus-login doctor` 检查配置和网络环境。",
            self.config.max_retries,
        )
        return False



"""Tests for configuration precedence and safety."""

from __future__ import annotations

import os
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

from campus_login_tool.config import (
    ConfigError,
    determine_config_path,
    resolve_config,
    write_default_config,
)


class ConfigTests(unittest.TestCase):
    def test_cli_values_override_env_and_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\n"
                "username = file_user\n"
                "password = file_pass\n"
                "\n[settings]\n"
                "check_url = https://file.example\n"
                "check_interval = 10\n"
                "max_retries = 2\n",
                encoding="utf-8",
            )
            config_path.chmod(0o600)

            with patch.dict(
                os.environ,
                {
                    "CAMPUS_LOGIN_USERNAME": "env_user",
                    "CAMPUS_LOGIN_PASSWORD": "env_pass",
                    "CAMPUS_LOGIN_CHECK_URL": "https://env.example",
                    "CAMPUS_LOGIN_INTERVAL": "20",
                    "CAMPUS_LOGIN_RETRIES": "4",
                },
                clear=False,
            ):
                resolved = resolve_config(
                    cli_values={
                        "username": "cli_user",
                        "password": "cli_pass",
                        "check_url": "https://cli.example",
                        "check_interval": 30,
                        "max_retries": 8,
                    },
                    config_path=config_path,
                )

        self.assertEqual(resolved.username, "cli_user")
        self.assertEqual(resolved.password, "cli_pass")
        self.assertEqual(resolved.check_url, "https://cli.example")
        self.assertEqual(resolved.check_interval, 30)
        self.assertEqual(resolved.max_retries, 8)
        self.assertEqual(resolved.username_source, "cli")
        self.assertEqual(resolved.password_source, "cli")

    def test_env_values_override_config_when_cli_missing(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\n"
                "username = file_user\n"
                "password = file_pass\n",
                encoding="utf-8",
            )
            config_path.chmod(0o600)

            with patch.dict(
                os.environ,
                {
                    "CAMPUS_LOGIN_USERNAME": "env_user",
                    "CAMPUS_LOGIN_PASSWORD": "env_pass",
                },
                clear=False,
            ):
                resolved = resolve_config(cli_values={}, config_path=config_path)

        self.assertEqual(resolved.username, "env_user")
        self.assertEqual(resolved.password, "env_pass")
        self.assertEqual(resolved.username_source, "env")
        self.assertEqual(resolved.password_source, "env")

    @unittest.skipIf(os.name == "nt", "POSIX file permission checks are not portable to Windows")
    def test_password_file_requires_private_permissions(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\nusername = user\npassword = secret\n",
                encoding="utf-8",
            )
            config_path.chmod(0o644)

            with self.assertRaises(ConfigError):
                resolve_config(cli_values={}, config_path=config_path)

    @unittest.skipIf(os.name == "nt", "POSIX file permission checks are not portable to Windows")
    def test_cli_password_does_not_fail_on_insecure_unused_config_password(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\nusername = file_user\npassword = file_secret\n",
                encoding="utf-8",
            )
            config_path.chmod(0o644)

            resolved = resolve_config(
                cli_values={"username": "cli_user", "password": "cli_secret"},
                config_path=config_path,
            )

        self.assertEqual(resolved.username, "cli_user")
        self.assertEqual(resolved.password, "cli_secret")

    def test_missing_credentials_raise_when_required(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text("[settings]\ncheck_interval = 30\n", encoding="utf-8")

            with self.assertRaises(ConfigError):
                resolve_config(cli_values={}, config_path=config_path)

    def test_write_default_config_creates_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            created = write_default_config(config_path)

            self.assertEqual(created, config_path)
            self.assertTrue(config_path.exists())

    def test_determine_config_path_prefers_cli(self) -> None:
        with patch.dict(os.environ, {"CAMPUS_LOGIN_CONFIG": "/tmp/from-env.conf"}, clear=False):
            resolved = determine_config_path("/tmp/from-cli.conf")

        self.assertEqual(str(resolved), "/tmp/from-cli.conf")



"""Configuration loading and validation."""

import os
import stat
from collections.abc import Mapping
from configparser import ConfigParser
from configparser import Error as ConfigParserError
from dataclasses import dataclass
from pathlib import Path
from typing import Any

DEFAULT_TRIGGER_URL = "http://172.16.128.139/"
DEFAULT_INTERNET_CHECK_URL = "https://www.baidu.com"
DEFAULT_CHECK_INTERVAL = 30
DEFAULT_MAX_RETRIES = 3
DEFAULT_CONFIG_PATH = Path.home() / ".campus_login.conf"

ENV_CONFIG_PATH = "CAMPUS_LOGIN_CONFIG"
ENV_USERNAME = "CAMPUS_LOGIN_USERNAME"
ENV_PASSWORD = "CAMPUS_LOGIN_PASSWORD"
ENV_CHECK_URL = "CAMPUS_LOGIN_CHECK_URL"
ENV_INTERVAL = "CAMPUS_LOGIN_INTERVAL"
ENV_RETRIES = "CAMPUS_LOGIN_RETRIES"

CONFIG_TEMPLATE = """# 校园网自动登录配置文件
# 推荐用 `campus-login init-config` 生成本文件。
# 凭证以明文保存，仅建议在可信设备上使用，并将权限限制为 600。

[credentials]
# username = your_username
# password = your_password

[settings]
# 用于检测是否已联网；默认值通常够用
check_url = https://www.baidu.com

# watch 模式下的检测间隔（秒）
check_interval = 30

# 登录失败时的最大重试次数
max_retries = 3
"""


class ConfigError(ValueError):
    """Raised when configuration is invalid or unsafe."""


@dataclass(slots=True)
class ResolvedConfig:
    """Fully resolved runtime configuration."""

    username: str
    password: str
    check_url: str
    check_interval: int
    max_retries: int
    trigger_url: str
    config_path: Path
    username_source: str
    password_source: str
    check_url_source: str
    interval_source: str
    retries_source: str

    @property
    def has_credentials(self) -> bool:
        return bool(self.username and self.password)


def determine_config_path(cli_path: str | None) -> Path:
    """Resolve the configuration path with CLI > env > default precedence."""
    raw_path = cli_path or os.getenv(ENV_CONFIG_PATH)
    return Path(raw_path).expanduser() if raw_path else DEFAULT_CONFIG_PATH


def _read_config_file(path: Path) -> dict[str, Any]:
    if not path.exists():
        return {}

    parser = ConfigParser()
    try:
        parser.read(path, encoding="utf-8")
    except ConfigParserError as exc:
        raise ConfigError(f"无法解析配置文件 {path}: {exc}") from exc

    credentials = parser["credentials"] if parser.has_section("credentials") else {}
    settings = parser["settings"] if parser.has_section("settings") else {}

    password = _normalize_value(credentials.get("password"))

    return {
        "username": _normalize_value(credentials.get("username")),
        "password": password,
        "check_url": _normalize_value(settings.get("check_url")),
        "check_interval": _parse_optional_int(
            settings.get("check_interval"),
            "配置文件中的 check_interval",
        ),
        "max_retries": _parse_optional_int(
            settings.get("max_retries"),
            "配置文件中的 max_retries",
        ),
    }


def _read_env() -> dict[str, Any]:
    return {
        "username": _normalize_value(os.getenv(ENV_USERNAME)),
        "password": _normalize_value(os.getenv(ENV_PASSWORD)),
        "check_url": _normalize_value(os.getenv(ENV_CHECK_URL)),
        "check_interval": _parse_optional_int(os.getenv(ENV_INTERVAL), ENV_INTERVAL),
        "max_retries": _parse_optional_int(os.getenv(ENV_RETRIES), ENV_RETRIES),
    }


def resolve_config(
    *,
    cli_values: Mapping[str, Any],
    config_path: Path,
    require_credentials: bool = True,
) -> ResolvedConfig:
    """Merge CLI, environment, config file, and defaults into a runtime config."""
    file_values = _read_config_file(config_path)
    env_values = _read_env()

    username, username_source = _first_defined(
        (cli_values.get("username"), "cli"),
        (env_values["username"], "env"),
        (file_values.get("username"), "config"),
        (None, "missing"),
    )
    password, password_source = _first_defined(
        (cli_values.get("password"), "cli"),
        (env_values["password"], "env"),
        (file_values.get("password"), "config"),
        (None, "missing"),
    )
    check_url, check_url_source = _first_defined(
        (cli_values.get("check_url"), "cli"),
        (env_values["check_url"], "env"),
        (file_values.get("check_url"), "config"),
        (DEFAULT_INTERNET_CHECK_URL, "default"),
    )
    check_interval, interval_source = _first_defined(
        (cli_values.get("check_interval"), "cli"),
        (env_values["check_interval"], "env"),
        (file_values.get("check_interval"), "config"),
        (DEFAULT_CHECK_INTERVAL, "default"),
    )
    max_retries, retries_source = _first_defined(
        (cli_values.get("max_retries"), "cli"),
        (env_values["max_retries"], "env"),
        (file_values.get("max_retries"), "config"),
        (DEFAULT_MAX_RETRIES, "default"),
    )

    if check_interval is None or check_interval <= 0:
        raise ConfigError("检测间隔必须是正整数。")
    if max_retries is None or max_retries <= 0:
        raise ConfigError("最大重试次数必须是正整数。")

    username = username or ""
    password = password or ""

    if require_credentials and not username:
        raise ConfigError("未找到校园网用户名。请使用 --username、环境变量或配置文件提供。")
    if require_credentials and not password:
        raise ConfigError(
            "未找到校园网密码。请通过交互输入、--password-stdin、环境变量或配置文件提供。"
        )

    if password_source == "config":
        ensure_secure_config_permissions(config_path)

    return ResolvedConfig(
        username=username,
        password=password,
        check_url=check_url,
        check_interval=check_interval,
        max_retries=max_retries,
        trigger_url=DEFAULT_TRIGGER_URL,
        config_path=config_path,
        username_source=username_source,
        password_source=password_source,
        check_url_source=check_url_source,
        interval_source=interval_source,
        retries_source=retries_source,
    )


def write_default_config(path: Path, force: bool = False) -> Path:
    """Create a starter configuration file with secure permissions."""
    path = path.expanduser()
    if path.exists() and not force:
        raise ConfigError(f"配置文件已存在: {path}。如需覆盖，请使用 --force。")

    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(CONFIG_TEMPLATE, encoding="utf-8")
    try:
        path.chmod(0o600)
    except OSError:
        pass
    return path


def ensure_secure_config_permissions(path: Path) -> None:
    """Reject password-bearing config files that are readable by others."""
    if os.name == "nt":
        return

    mode = stat.S_IMODE(path.stat().st_mode)
    if mode & 0o077:
        chmod_hint = f"chmod 600 {path}"
        raise ConfigError(
            f"配置文件权限过宽 ({oct(mode)})：{path}。"
            f"如果文件中保存了密码，请先执行 `{chmod_hint}`。"
        )


def describe_credential_source(resolved: ResolvedConfig) -> str:
    """Return a user-facing description of where credentials came from."""
    if resolved.password_source == "cli":
        return "CLI"
    if resolved.password_source == "env":
        return "环境变量"
    if resolved.password_source == "config":
        return f"配置文件 ({resolved.config_path})"
    return "未配置"


def _parse_optional_int(raw_value: str | None, label: str) -> int | None:
    value = _normalize_value(raw_value)
    if value is None:
        return None
    try:
        return int(value)
    except (TypeError, ValueError) as exc:
        raise ConfigError(f"{label} 必须是整数，当前值为 {value!r}。") from exc


def _first_defined(*candidates: tuple[Any, str]) -> tuple[Any, str]:
    for value, source in candidates:
        if value is None:
            continue
        return value, source
    return None, "missing"


def _normalize_value(value: str | None) -> str | None:
    if value is None:
        return None
    stripped = value.strip()
    return stripped or None



"""Logging helpers for the campus login CLI."""

import logging
from pathlib import Path
from urllib.parse import urlparse


def setup_logging(verbose: bool = False, log_file: str | None = None) -> logging.Logger:
    """Create the application logger."""
    logger = logging.getLogger("campus_login")
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)
    logger.handlers.clear()
    logger.propagate = False

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG if verbose else logging.INFO)
    console_handler.setFormatter(
        logging.Formatter("%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    )
    logger.addHandler(console_handler)

    if log_file:
        log_path = Path(log_file).expanduser()
        log_path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(log_path, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s [%(levelname)s] %(name)s - %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )
        logger.addHandler(file_handler)

    return logger


def mask_value(value: str, keep_start: int = 2, keep_end: int = 2) -> str:
    """Return a minimally identifying masked value for logs."""
    if not value:
        return "<empty>"

    if len(value) <= keep_start + keep_end:
        return "*" * len(value)

    hidden = "*" * (len(value) - keep_start - keep_end)
    return f"{value[:keep_start]}{hidden}{value[-keep_end:]}"


def sanitize_url(url: str) -> str:
    """Drop query fragments before logging URLs."""
    parsed = urlparse(url)
    if not parsed.scheme or not parsed.netloc:
        return url
    return f"{parsed.scheme}://{parsed.netloc}{parsed.path}"


def truncate_text(content: str, limit: int = 160) -> str:
    """Shorten long debug strings to avoid noisy logs."""
    if len(content) <= limit:
        return content
    return f"{content[:limit]}..."



"""Tests for low-level protocol behavior."""

import gzip
import json
import logging
import unittest
from pathlib import Path
from unittest.mock import Mock, patch
from urllib.parse import urlparse

import requests

from campus_login_tool.client import CampusLoginClient
from campus_login_tool.config import ResolvedConfig

FIXTURES = Path(__file__).parent / "fixtures"


class FakeResponse:
    def __init__(
        self,
        *,
        content: bytes,
        status_code: int = 200,
        headers=None,
        json_data=None,
        json_exc=None,
        url: str = "http://172.16.128.139/",
        history=None,
    ):
        self.content = content
        self.status_code = status_code
        self.headers = headers or {}
        self._json_data = json_data
        self._json_exc = json_exc
        self.url = url
        self.history = list(history or [])

    def json(self):
        if self._json_exc is not None:
            raise self._json_exc
        return self._json_data

    def raise_for_status(self) -> None:
        if self.status_code >= 400:
            raise requests.HTTPError(f"HTTP {self.status_code}")


class FakeSession:
    def __init__(self, get_responses=None, post_responses=None):
        self.get_responses = list(get_responses or [])
        self.post_responses = list(post_responses or [])
        self.headers = {}

    def get(self, *args, **kwargs):
        return self.get_responses.pop(0)

    def post(self, *args, **kwargs):
        return self.post_responses.pop(0)

    def close(self) -> None:
        return None


def make_config() -> ResolvedConfig:
    return ResolvedConfig(
        username="test_user",
        password="test_password",
        check_url="https://check.example",
        check_interval=30,
        max_retries=3,
        trigger_url="http://172.16.128.139/",
        config_path=Path("/tmp/campus.conf"),
        username_source="cli",
        password_source="cli",
        check_url_source="default",
        interval_source="default",
        retries_source="default",
    )


def make_logger() -> logging.Logger:
    logger = logging.getLogger("campus_login_test")
    logger.handlers.clear()
    logger.addHandler(logging.NullHandler())
    logger.propagate = False
    return logger


class ClientTests(unittest.TestCase):
    def test_decode_response_supports_gzip_and_gbk(self) -> None:
        text = "登录页面"
        response = FakeResponse(
            content=gzip.compress(text.encode("gb18030")),
            headers={"Content-Encoding": "gzip"},
        )
        client = CampusLoginClient(make_config(), make_logger())

        self.assertEqual(client._decode_response(response), text)

    def test_extract_redirect_url_from_fixture(self) -> None:
        html = (FIXTURES / "redirect.html").read_text(encoding="utf-8")
        redirect_url = CampusLoginClient._extract_redirect_url(html)

        self.assertIsNotNone(redirect_url)
        self.assertIn("/eportal/index.jsp", redirect_url)

    def test_parse_login_response_success_and_failure(self) -> None:
        success, message = CampusLoginClient._parse_login_response(
            '{"result":"success","message":""}'
        )
        failure, fail_message = CampusLoginClient._parse_login_response(
            '{"result":"fail","message":"用户名或密码错误"}'
        )

        self.assertTrue(success)
        self.assertEqual(message, "登录成功")
        self.assertFalse(failure)
        self.assertEqual(fail_message, "用户名或密码错误")

    def test_build_page_info_request_encodes_query_once(self) -> None:
        client = CampusLoginClient(make_config(), make_logger())
        login_url = CampusLoginClient._extract_redirect_url(
            (FIXTURES / "redirect.html").read_text(encoding="utf-8")
        )
        query = urlparse(login_url).query

        prepared = client.build_page_info_request(login_url, query)

        self.assertIn("queryString", prepared.data)
        self.assertIn("%3D", prepared.data["queryString"])
        self.assertEqual(prepared.headers["Origin"], "http://172.16.128.139")

    def test_build_login_request_contains_expected_payload(self) -> None:
        client = CampusLoginClient(make_config(), make_logger())
        login_url = CampusLoginClient._extract_redirect_url(
            (FIXTURES / "redirect.html").read_text(encoding="utf-8")
        )
        query = urlparse(login_url).query

        prepared = client.build_login_request(login_url, query, "encrypted", "campus")

        self.assertEqual(prepared.data["userId"], "test_user")
        self.assertEqual(prepared.data["password"], "encrypted")
        self.assertEqual(prepared.data["service"], "campus")
        self.assertIn("%3D", prepared.data["queryString"])

    def test_build_interface_request_targets_portal_method(self) -> None:
        client = CampusLoginClient(make_config(), make_logger())

        prepared = client.build_interface_request("logout", {"userIndex": "abc"})

        self.assertEqual(prepared.url, "http://172.16.128.139/eportal/InterFace.do?method=logout")
        self.assertEqual(prepared.data["userIndex"], "abc")
        self.assertEqual(prepared.headers["Origin"], "http://172.16.128.139")

    def test_get_page_info_falls_back_to_manual_json_decode(self) -> None:
        json_content = (FIXTURES / "pageinfo.json").read_text(encoding="utf-8")
        client = CampusLoginClient(make_config(), make_logger(), session_factory=FakeSession)
        client.session = FakeSession(
            post_responses=[
                FakeResponse(
                    content=json_content.encode("utf-8"),
                    json_exc=json.JSONDecodeError("bad", json_content, 0),
                )
            ]
        )
        login_url = "http://172.16.128.139/eportal/index.jsp?foo=bar"

        payload = client._get_page_info(login_url, "foo=bar")

        self.assertEqual(payload["publicKeyExponent"], "10001")
        self.assertIn("publicKeyModulus", payload)

    def test_parse_json_response_uses_custom_decoder_for_gbk_content(self) -> None:
        payload_text = '{"result":"fail","message":"当前用户不在线，请重新认证"}'
        response = FakeResponse(
            content=payload_text.encode("gb18030"),
            json_data={"result": "fail", "message": "ç¨æ·ä¿¡æ¯ä¸å®æ´ï¼è¯·ç¨åéè¯"},
        )
        client = CampusLoginClient(make_config(), make_logger())

        payload = client._parse_json_response(response, action="getOnlineUserInfo")

        self.assertEqual(payload["message"], "当前用户不在线，请重新认证")

    def test_login_uses_mac_value_for_password_encryption(self) -> None:
        redirect_html = (FIXTURES / "redirect.html").read_text(encoding="utf-8").encode("utf-8")
        session = FakeSession(
            get_responses=[
                FakeResponse(content=redirect_html),
                FakeResponse(content=b"<html>login page</html>"),
            ]
        )
        client = CampusLoginClient(make_config(), make_logger(), session_factory=lambda: session)
        page_info = json.loads((FIXTURES / "pageinfo.json").read_text(encoding="utf-8"))

        with (
            patch(
                "campus_login_tool.client.encryptPassword",
                return_value="encrypted",
            ) as encrypt_mock,
            patch.object(
                client,
                "_get_page_info",
                return_value=page_info,
            ),
            patch.object(client, "_submit_login", return_value=(True, "ok")),
        ):
            result = client.login()

        self.assertTrue(result)
        encrypt_mock.assert_called_once()
        self.assertEqual(encrypt_mock.call_args.args[0], "test_password")
        self.assertEqual(encrypt_mock.call_args.args[3], "5fdcd87b4a9fecfe22921ef5c80e86b4")

    def test_login_caches_user_index_from_success_response(self) -> None:
        redirect_html = (FIXTURES / "redirect.html").read_text(encoding="utf-8").encode("utf-8")
        session = FakeSession(
            get_responses=[
                FakeResponse(content=redirect_html),
                FakeResponse(content=b"<html>login page</html>"),
            ]
        )
        client = CampusLoginClient(make_config(), make_logger(), session_factory=lambda: session)
        page_info = json.loads((FIXTURES / "pageinfo.json").read_text(encoding="utf-8"))

        with (
            patch("campus_login_tool.client.encryptPassword", return_value="encrypted"),
            patch.object(client, "_get_page_info", return_value=page_info),
            patch.object(
                client.session if client.session is not None else session,
                "post",
                return_value=FakeResponse(
                    content=b'{"result":"success","message":"","userIndex":"derived-index"}',
                    json_data={"result": "success", "message": "", "userIndex": "derived-index"},
                ),
            ),
        ):
            result = client.login()

        self.assertTrue(result)
        self.assertEqual(client._last_user_index, "derived-index")

    def test_logout_returns_success_when_portal_shows_login_page(self) -> None:
        redirect_html = (FIXTURES / "redirect.html").read_text(encoding="utf-8").encode("utf-8")
        session = FakeSession(
            get_responses=[
                FakeResponse(
                    content=redirect_html,
                    url="http://172.16.128.139/",
                )
            ]
        )
        client = CampusLoginClient(make_config(), make_logger(), session_factory=lambda: session)

        result = client.logout()

        self.assertTrue(result)

    def test_logout_posts_logout_after_online_check(self) -> None:
        session = FakeSession(
            get_responses=[
                FakeResponse(
                    content=b"",
                    url="http://172.16.128.139/eportal/success.jsp?userIndex=user-index",
                )
            ]
        )
        client = CampusLoginClient(make_config(), make_logger(), session_factory=lambda: session)

        with (
            patch.object(
                client,
                "get_online_user_info",
                return_value={"result": "success", "userId": "test_user"},
            ) as info_mock,
            patch.object(
                client,
                "_post_interface_json",
                return_value={"result": "success", "message": "下线成功"},
            ) as post_mock,
        ):
            result = client.logout()

        self.assertTrue(result)
        info_mock.assert_called_once_with("user-index")
        post_mock.assert_called_once_with("logout", {"userIndex": "user-index"})

    def test_logout_retries_when_online_user_info_is_incomplete(self) -> None:
        session = FakeSession(
            get_responses=[
                FakeResponse(
                    content=b"",
                    url="http://172.16.128.139/eportal/success.jsp?userIndex=user-index",
                ),
                FakeResponse(
                    content=b"",
                    url="http://172.16.128.139/eportal/success.jsp?userIndex=user-index-2",
                ),
            ]
        )
        sleep_calls = []
        client = CampusLoginClient(
            make_config(),
            make_logger(),
            session_factory=lambda: session,
            sleep_func=sleep_calls.append,
        )

        with (
            patch.object(
                client,
                "get_online_user_info",
                side_effect=[
                    {"result": "fail", "message": "用户信息不完整，请稍后重试"},
                    {"result": "success", "userId": "test_user"},
                ],
            ) as info_mock,
            patch.object(
                client,
                "_post_interface_json",
                return_value={"result": "success", "message": "下线成功"},
            ) as post_mock,
        ):
            result = client.logout()

        self.assertTrue(result)
        self.assertEqual(sleep_calls, [1])
        self.assertEqual(info_mock.call_args_list[0].args[0], "user-index")
        self.assertEqual(info_mock.call_args_list[1].args[0], "user-index-2")
        post_mock.assert_called_once_with("logout", {"userIndex": "user-index-2"})

    def test_logout_retry_treats_not_online_as_success(self) -> None:
        redirect_html = (FIXTURES / "redirect.html").read_text(encoding="utf-8").encode("utf-8")
        session = FakeSession(
            get_responses=[
                FakeResponse(
                    content=b"",
                    url="http://172.16.128.139/eportal/success.jsp?userIndex=user-index",
                ),
                FakeResponse(
                    content=redirect_html,
                    url="http://172.16.128.139/",
                ),
            ]
        )
        sleep_calls = []
        client = CampusLoginClient(
            make_config(),
            make_logger(),
            session_factory=lambda: session,
            sleep_func=sleep_calls.append,
        )

        with patch.object(
            client,
            "get_online_user_info",
            return_value={"result": "fail", "message": "用户信息不完整，请稍后重试"},
        ):
            result = client.logout()

        self.assertTrue(result)
        self.assertEqual(sleep_calls, [1])

    def test_find_login_page_url_supports_direct_final_url(self) -> None:
        client = CampusLoginClient(make_config(), make_logger())
        response = FakeResponse(
            content=b"<html>login page</html>",
            url=(
                "http://172.16.128.139/eportal/index.jsp?"
                "wlanuserip=1&nasip=2&mac=3"
            ),
        )

        login_url = client._find_login_page_url(response)

        self.assertEqual(
            login_url,
            "http://172.16.128.139/eportal/index.jsp?wlanuserip=1&nasip=2&mac=3",
        )

    def test_login_with_retry_uses_backoff(self) -> None:
        sleep_calls = []
        client = CampusLoginClient(make_config(), make_logger(), sleep_func=sleep_calls.append)
        client.login = Mock(side_effect=[False, False, True])
        client.check_internet = Mock(return_value=True)

        result = client.login_with_retry()

        self.assertTrue(result)
        self.assertEqual(sleep_calls, [2, 4, 2])



"""Command line interface for the campus login tool."""

import argparse
import sys
from collections.abc import Iterable

from . import __version__
from .client import CampusLoginClient
from .config import (
    DEFAULT_CONFIG_PATH,
    ConfigError,
    ResolvedConfig,
    describe_credential_source,
    determine_config_path,
    resolve_config,
    write_default_config,
)
from .logging_utils import setup_logging
from .watch import WatchRunner


def build_parser() -> argparse.ArgumentParser:
    """Build the primary CLI parser."""
    parser = argparse.ArgumentParser(
        prog="campus-login",
        description="校园网自动登录工具（Eportal 动态公钥型）",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    common = argparse.ArgumentParser(add_help=False)
    common.add_argument("-v", "--verbose", action="store_true", help="输出调试日志")
    common.add_argument("--log-file", help="日志文件路径")

    runtime = argparse.ArgumentParser(add_help=False)
    runtime.add_argument("--config", help=f"配置文件路径，默认 {DEFAULT_CONFIG_PATH}")
    runtime.add_argument("-u", "--username", help="校园网用户名")
    runtime.add_argument(
        "-p",
        "--password",
        help="已弃用：明文密码参数，建议改用交互输入或 --password-stdin",
    )
    runtime.add_argument(
        "--password-stdin",
        action="store_true",
        help="从标准输入读取密码，适合与密码管理器或管道配合使用",
    )
    runtime.add_argument("--check-url", dest="check_url", help="网络检测 URL")
    runtime.add_argument("--interval", dest="check_interval", type=int, help="检测间隔（秒）")
    runtime.add_argument("--retries", dest="max_retries", type=int, help="最大重试次数")

    subparsers = parser.add_subparsers(dest="command", required=True)

    login_parser = subparsers.add_parser(
        "login",
        parents=[common, runtime],
        help="执行一次登录",
    )
    login_parser.add_argument("--force", action="store_true", help="跳过登录前的联网检测")
    login_parser.set_defaults(handler=handle_login)

    logout_parser = subparsers.add_parser(
        "logout",
        parents=[common],
        help="退出当前校园网会话",
    )
    logout_parser.add_argument("--config", help=f"配置文件路径，默认 {DEFAULT_CONFIG_PATH}")
    logout_parser.add_argument("--user-index", help="直接指定 userIndex，跳过本地自动发现")
    logout_parser.set_defaults(handler=handle_logout)

    watch_parser = subparsers.add_parser(
        "watch",
        parents=[common, runtime],
        help="持续监控网络并在断网时自动重连",
    )
    watch_parser.add_argument("--force", action="store_true", help="启动时跳过联网检测")
    watch_parser.set_defaults(handler=handle_watch)

    doctor_parser = subparsers.add_parser(
        "doctor",
        parents=[common, runtime],
        help="检查配置、安全性和当前网络环境",
    )
    doctor_parser.add_argument("--timeout", type=int, default=5, help="网络检测超时（秒）")
    doctor_parser.set_defaults(handler=handle_doctor)

    init_parser = subparsers.add_parser(
        "init-config",
        parents=[common],
        help="生成示例配置文件",
    )
    init_parser.add_argument("--config", help=f"输出路径，默认 {DEFAULT_CONFIG_PATH}")
    init_parser.add_argument("--force", action="store_true", help="覆盖已有配置文件")
    init_parser.set_defaults(handler=handle_init_config)

    return parser


def build_legacy_parser() -> argparse.ArgumentParser:
    """Build the compatibility parser for the historical script interface."""
    parser = argparse.ArgumentParser(
        prog="campus_login.py",
        description="校园网自动登录脚本（兼容模式，建议改用 `campus-login`）",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
兼容示例:
  python campus_login.py -u myuser
  python campus_login.py -u myuser --password-stdin
  python campus_login.py --daemon

推荐改用:
  campus-login login
  campus-login watch
""",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    parser.add_argument("-u", "--username", help="校园网用户名")
    parser.add_argument(
        "-p",
        "--password",
        help="已弃用：明文密码参数，建议改用交互输入或 --password-stdin",
    )
    parser.add_argument(
        "--password-stdin",
        action="store_true",
        help="从标准输入读取密码，适合与密码管理器或管道配合使用",
    )
    parser.add_argument(
        "-d",
        "--daemon",
        action="store_true",
        help="已弃用：改用 `campus-login watch`",
    )
    parser.add_argument("--config", help=f"配置文件路径，默认 {DEFAULT_CONFIG_PATH}")
    parser.add_argument("--check-url", dest="check_url", help="网络检测 URL")
    parser.add_argument("-i", "--interval", dest="check_interval", type=int, help="检测间隔（秒）")
    parser.add_argument("-r", "--retries", dest="max_retries", type=int, help="最大重试次数")
    parser.add_argument("--force", action="store_true", help="跳过登录前的联网检测")
    parser.add_argument("-v", "--verbose", action="store_true", help="输出调试日志")
    parser.add_argument("--log-file", help="日志文件路径")
    return parser


def main(argv: Iterable[str] | None = None) -> int:
    """Run the primary CLI."""
    parser = build_parser()
    args = parser.parse_args(list(argv) if argv is not None else None)
    logger = setup_logging(args.verbose, args.log_file)
    try:
        return int(args.handler(args, logger))
    except ConfigError as exc:
        logger.error(str(exc))
        return 2


def legacy_main(argv: Iterable[str] | None = None) -> int:
    """Run the historical single-command interface."""
    argv_list = list(argv) if argv is not None else sys.argv[1:]
    if argv_list and argv_list[0] in {"login", "logout", "watch", "doctor", "init-config"}:
        return main(argv_list)

    parser = build_legacy_parser()
    args = parser.parse_args(argv_list)
    logger = setup_logging(args.verbose, args.log_file)
    logger.warning("兼容模式：`python campus_login.py` 已弃用，请改用 `campus-login`。")
    if args.daemon:
        logger.warning("兼容模式：`--daemon` 已弃用，请改用 `campus-login watch`。")

    try:
        if args.daemon:
            return handle_watch(args, logger)
        return handle_login(args, logger)
    except ConfigError as exc:
        logger.error(str(exc))
        return 2


def handle_login(args: argparse.Namespace, logger) -> int:
    """Execute a single login attempt."""
    if args.password is not None:
        logger.warning("`--password` 已弃用，建议改用交互输入、环境变量或 `--password-stdin`。")

    config = _resolve_runtime_config(args, require_credentials=True)
    _log_config_summary(config, logger)
    client = CampusLoginClient(config, logger)

    if not args.force:
        logger.info("正在检测网络连通性: %s", config.check_url)
        if client.check_internet():
            logger.info("网络已连接，无需登录。")
            return 0

    return 0 if client.login_with_retry() else 1


def handle_watch(args: argparse.Namespace, logger) -> int:
    """Run the foreground watch loop."""
    if getattr(args, "password", None) is not None:
        logger.warning("`--password` 已弃用，建议改用交互输入、环境变量或 `--password-stdin`。")

    config = _resolve_runtime_config(args, require_credentials=True)
    _log_config_summary(config, logger)
    client = CampusLoginClient(config, logger)

    if not args.force:
        logger.info("正在检测网络连通性: %s", config.check_url)
        if client.check_internet():
            logger.info("网络已连接，watch 模式将继续监控。")
        else:
            logger.info("当前未联网，watch 模式会先尝试登录。")
            client.login_with_retry()

    runner = WatchRunner(client, logger)
    runner.run()
    return 0


def handle_logout(args: argparse.Namespace, logger) -> int:
    """Log out the current online session if one exists."""
    config = _resolve_runtime_config(args, require_credentials=False)
    _log_config_summary(config, logger, include_credentials=False)
    client = CampusLoginClient(config, logger)
    return 0 if client.logout(user_index=getattr(args, "user_index", None)) else 1


def handle_doctor(args: argparse.Namespace, logger) -> int:
    """Run configuration and environment diagnostics."""
    config = _resolve_runtime_config(args, require_credentials=False)
    _log_config_summary(config, logger, include_credentials=False)

    issues = 0
    config_exists = config.config_path.exists()
    if config_exists:
        logger.info("配置文件: %s", config.config_path)
    else:
        logger.info("配置文件未找到: %s", config.config_path)

    if config.username:
        logger.info("用户名来源: %s", config.username_source)
    else:
        logger.warning("未检测到用户名，请配置 --username、环境变量或配置文件。")
        issues += 1

    if config.password:
        logger.info("密码来源: %s", describe_credential_source(config))
    else:
        logger.warning("未检测到密码，请使用交互输入、--password-stdin、环境变量或配置文件。")
        issues += 1

    client = CampusLoginClient(config, logger)
    logger.info("正在检查互联网探测地址: %s", config.check_url)
    if client.check_internet(timeout=args.timeout):
        logger.info("互联网探测通过，当前看起来已经联网。")
    else:
        logger.warning("互联网探测未通过，这通常表示尚未登录，或当前不在校园网环境。")
        issues += 1

    logger.info("正在检查认证触发地址")
    ok, detail = client.probe_auth_trigger(timeout=args.timeout)
    if ok:
        logger.info("认证触发检查通过: %s", detail)
    else:
        logger.warning("认证触发检查失败: %s", detail)
        issues += 1

    if issues == 0:
        logger.info("doctor 检查完成，未发现明显问题。")
        return 0

    logger.warning("doctor 检查完成，发现 %s 个需要关注的问题。", issues)
    return 1


def handle_init_config(args: argparse.Namespace, logger) -> int:
    """Create a starter config file."""
    config_path = determine_config_path(args.config)
    created_path = write_default_config(config_path, force=args.force)
    logger.info("已生成配置文件模板: %s", created_path)
    logger.info("如需保存密码，请确认文件权限保持为 600。")
    return 0


def _resolve_runtime_config(
    args: argparse.Namespace,
    *,
    require_credentials: bool,
) -> ResolvedConfig:
    config_path = determine_config_path(getattr(args, "config", None))
    if getattr(args, "password", None) and getattr(args, "password_stdin", False):
        raise ConfigError("`--password` 和 `--password-stdin` 不能同时使用。")

    password_from_stdin = None
    if getattr(args, "password_stdin", False):
        password_from_stdin = sys.stdin.read().rstrip("\r\n")
        if not password_from_stdin:
            raise ConfigError("`--password-stdin` 已启用，但标准输入中没有读取到密码。")

    cli_values = {
        "username": getattr(args, "username", None),
        "password": (
            password_from_stdin
            if password_from_stdin is not None
            else getattr(args, "password", None)
        ),
        "check_url": getattr(args, "check_url", None),
        "check_interval": getattr(args, "check_interval", None),
        "max_retries": getattr(args, "max_retries", None),
    }

    resolved = resolve_config(
        cli_values=cli_values,
        config_path=config_path,
        require_credentials=False,
    )

    if require_credentials and not resolved.username:
        prompted_username = _read_username_from_tty_if_needed()
        if prompted_username:
            cli_values["username"] = prompted_username

    if require_credentials and not resolved.password and password_from_stdin is None:
        prompted_password = _read_password_from_tty_if_needed()
        if prompted_password:
            cli_values["password"] = prompted_password

    return resolve_config(
        cli_values=cli_values,
        config_path=config_path,
        require_credentials=require_credentials,
    )


def _read_username_from_tty_if_needed() -> str | None:
    if not sys.stdin.isatty():
        return None
    username = input("请输入校园网用户名: ")
    return username.strip() or None


def _read_password_from_tty_if_needed() -> str | None:
    if not sys.stdin.isatty():
        return None

    import getpass

    password = getpass.getpass("请输入校园网密码: ")
    return password.strip() or None


def _log_config_summary(
    config: ResolvedConfig,
    logger,
    *,
    include_credentials: bool = True,
) -> None:
    logger.info("配置文件路径: %s", config.config_path)
    logger.info(
        "配置来源: check_url=%s, interval=%s, retries=%s",
        config.check_url_source,
        config.interval_source,
        config.retries_source,
    )
    if include_credentials:
        logger.info("密码来源: %s", describe_credential_source(config))



"""RSA helpers for the captive portal login flow.

The portal's JavaScript uses a custom BigInt implementation, but the actual
algorithm is plain modular exponentiation over 16-bit words. Python's built-in
integers can perform the same work directly and much more efficiently.
"""


biRadixBits = 16
bitsPerDigit = biRadixBits
biRadix = 1 << biRadixBits
maxDigitVal = biRadix - 1


def _hex_words(value: str) -> list[int]:
    """Split a hex string into JS-compatible 16-bit words, least significant first."""
    words = []
    for end in range(len(value), 0, -4):
        start = max(end - 4, 0)
        words.append(int(value[start:end], 16))
    while len(words) > 1 and words[-1] == 0:
        words.pop()
    return words


def _chunk_size_from_modulus(modulus_hex: str) -> int:
    """Match the historical `2 * biHighIndex(modulus)` chunk sizing rule."""
    words = _hex_words(modulus_hex)
    high_index = len(words) - 1
    chunk_size = 2 * high_index
    if chunk_size <= 0:
        raise ValueError("RSA modulus is too short to determine a valid chunk size.")
    return chunk_size


def _encode_plaintext_block(chars: list[int]) -> int:
    """Pack plaintext into 16-bit little-endian words like the portal JS."""
    block = 0
    shift = 0
    for index in range(0, len(chars), 2):
        word = chars[index]
        if index + 1 < len(chars):
            word |= chars[index + 1] << 8
        block |= word << shift
        shift += biRadixBits
    return block


def _int_to_portal_hex(value: int) -> str:
    """Format integers the same way the JS BigInt `biToHex` helper does."""
    if value == 0:
        return "0000"

    groups = []
    while value:
        groups.append(f"{value & maxDigitVal:04x}")
        value >>= biRadixBits
    return "".join(reversed(groups))


def encryptPassword(
    password: str,
    publicKeyExponent: str,
    publicKeyModulus: str,
    macString: str = "111111111",
) -> str:
    """Encrypt the password exactly as the captive portal expects."""
    password_mac = f"{password}>{macString}"
    password_encode = password_mac[::-1]
    chunk_size = _chunk_size_from_modulus(publicKeyModulus)

    plaintext = [ord(char) for char in password_encode]
    while len(plaintext) % chunk_size != 0:
        plaintext.append(0)

    exponent = int(publicKeyExponent, 16)
    modulus = int(publicKeyModulus, 16)
    encrypted_blocks = []

    for offset in range(0, len(plaintext), chunk_size):
        block = _encode_plaintext_block(plaintext[offset : offset + chunk_size])
        encrypted = pow(block, exponent, modulus)
        encrypted_blocks.append(_int_to_portal_hex(encrypted))

    return " ".join(encrypted_blocks)



"""CLI smoke tests."""

import io
import tempfile
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from campus_login_tool.cli import legacy_main, main


class CliTests(unittest.TestCase):
    def test_main_help_exits_cleanly(self) -> None:
        with self.assertRaises(SystemExit) as ctx:
            main(["--help"])

        self.assertEqual(ctx.exception.code, 0)

    def test_init_config_creates_template(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            path = Path(tmpdir) / "campus.conf"
            exit_code = main(["init-config", "--config", str(path)])

            self.assertEqual(exit_code, 0)
            self.assertTrue(path.exists())

    def test_doctor_smoke(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\nusername = user\npassword = secret\n",
                encoding="utf-8",
            )
            config_path.chmod(0o600)

            with (
                patch(
                    "campus_login_tool.cli.CampusLoginClient.check_internet",
                    return_value=True,
                ),
                patch(
                    "campus_login_tool.cli.CampusLoginClient.probe_auth_trigger",
                    return_value=(True, "http://172.16.128.139/eportal/index.jsp"),
                ),
            ):
                exit_code = main(["doctor", "--config", str(config_path)])

        self.assertEqual(exit_code, 0)

    def test_watch_smoke(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"
            config_path.write_text(
                "[credentials]\nusername = user\npassword = secret\n",
                encoding="utf-8",
            )
            config_path.chmod(0o600)

            with (
                patch(
                    "campus_login_tool.cli.CampusLoginClient.check_internet",
                    return_value=True,
                ),
                patch(
                    "campus_login_tool.cli.WatchRunner.run",
                    return_value=None,
                ) as run_mock,
            ):
                exit_code = main(["watch", "--config", str(config_path)])

        self.assertEqual(exit_code, 0)
        run_mock.assert_called_once()

    def test_logout_smoke_without_credentials(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"

            with patch(
                "campus_login_tool.cli.CampusLoginClient.logout",
                return_value=True,
            ) as logout_mock:
                exit_code = main(["logout", "--config", str(config_path)])

        self.assertEqual(exit_code, 0)
        logout_mock.assert_called_once_with(user_index=None)

    def test_legacy_mode_warns_about_deprecated_flags(self) -> None:
        mock_logger = Mock()
        with patch("campus_login_tool.cli.setup_logging", return_value=mock_logger), patch(
            "campus_login_tool.cli.WatchRunner.run",
            return_value=None,
        ):
            exit_code = legacy_main(
                ["-u", "user", "-p", "secret", "--daemon", "--force"]
            )

        self.assertEqual(exit_code, 0)
        warning_messages = "\n".join(call.args[0] for call in mock_logger.warning.call_args_list)
        self.assertIn("兼容模式：`python campus_login.py` 已弃用", warning_messages)
        self.assertIn("`--password` 已弃用", warning_messages)
        self.assertIn("`--daemon` 已弃用", warning_messages)

    def test_password_stdin_is_supported(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            config_path = Path(tmpdir) / "campus.conf"

            with patch(
                "campus_login_tool.cli.CampusLoginClient.login_with_retry",
                return_value=True,
            ):
                with patch("sys.stdin", io.StringIO("secret\n")):
                    exit_code = main(
                        [
                            "login",
                            "--config",
                            str(config_path),
                            "--username",
                            "user",
                            "--password-stdin",
                            "--force",
                        ]
                    )

        self.assertEqual(exit_code, 0)



"""Compatibility and documentation regression tests."""

import importlib
import unittest
from pathlib import Path

from campus_login_tool import security as package_security
from campus_login_tool.config import CONFIG_TEMPLATE

ROOT = Path(__file__).resolve().parent.parent


class CompatibilityTests(unittest.TestCase):
    def test_example_config_matches_generated_template(self) -> None:
        example_path = ROOT / "campus_login.conf.example"

        self.assertEqual(example_path.read_text(encoding="utf-8"), CONFIG_TEMPLATE)

    def test_root_security_module_re_exports_package_helpers(self) -> None:
        legacy_security = importlib.import_module("security")

        self.assertIs(legacy_security.encryptPassword, package_security.encryptPassword)
        self.assertEqual(legacy_security.biRadix, package_security.biRadix)



"""Foreground watch loop for reconnecting the campus network."""

import logging
import signal
import time
from collections.abc import Callable

from .client import CampusLoginClient

SleepFunc = Callable[[float], None]


class WatchRunner:
    """Continuously monitor network state and reconnect when needed."""

    def __init__(
        self,
        client: CampusLoginClient,
        logger: logging.Logger,
        *,
        sleep_func: SleepFunc = time.sleep,
        install_signal_handlers: bool = True,
    ) -> None:
        self.client = client
        self.logger = logger
        self.running = True
        self._sleep = sleep_func

        if install_signal_handlers:
            signal.signal(signal.SIGINT, self._signal_handler)
            signal.signal(signal.SIGTERM, self._signal_handler)

    def _signal_handler(self, signum, _frame) -> None:
        self.logger.info("收到退出信号 %s，正在停止 watch 模式。", signum)
        self.running = False

    def run(self) -> None:
        self.logger.info("watch 模式已启动")
        self.logger.info("检测间隔: %s 秒", self.client.config.check_interval)
        self.logger.info("检测 URL: %s", self.client.config.check_url)

        consecutive_failures = 0
        while self.running:
            try:
                if self.client.check_internet():
                    consecutive_failures = 0
                    self.logger.debug("网络正常")
                else:
                    consecutive_failures += 1
                    self.logger.warning("网络检测失败（连续 %s 次）", consecutive_failures)
                    if consecutive_failures >= 2:
                        self.logger.info("检测到网络断开，尝试重新登录")
                        if self.client.login_with_retry():
                            consecutive_failures = 0
                            self.logger.info("重新登录成功")
                        else:
                            self.logger.error("重新登录失败，将继续监控")

                for _ in range(self.client.config.check_interval):
                    if not self.running:
                        break
                    self._sleep(1)
            except Exception as exc:  # pragma: no cover - protection for long-running mode
                self.logger.exception("watch 模式发生异常: %s", exc)
                self._sleep(10)

        self.logger.info("watch 模式已退出")



"""Regression tests for RSA password encryption."""

import unittest

from campus_login_tool.security import encryptPassword

MODULUS = (
    "94dd2a8675fb779e6b9f7103698634cd400f27a154afa67af6166a43fc264172"
    "22a79506d34cacc7641946abda1785b7acf9910ad6a0978c91ec84d40b71d289"
    "1379af19ffb333e7517e390bd26ac312fe940c340466b4a5d4af1d65c3b59440"
    "78f96a1a51a5a53e4bc302818b7c9f63c4a1b07bd7d874cef1c3d4b2f5eb7871"
)


class SecurityTests(unittest.TestCase):
    def test_encrypt_password_matches_single_block_reference(self) -> None:
        encrypted = encryptPassword(
            "test_password",
            "10001",
            MODULUS,
            "5fdcd87b4a9fecfe22921ef5c80e86b4",
        )

        self.assertEqual(
            encrypted,
            "8df8fdd4d6f5a12fdd2119fb1e8a395c85b2fa69b4989ab688610e83d6a8fa86"
            "e748e011bebf85f77cac7806e433d29bba8390339be9c09a4089a03483ca2d5f"
            "ff6eeea31258d1047a048ef1589caf2c10c92d36eb76658d675eff630e3ee516"
            "dcb522b1ccc1168128a870d44f9d50445af52ad8a77b7dd45f361e159f4c6849",
        )

    def test_encrypt_password_matches_multi_block_reference(self) -> None:
        encrypted = encryptPassword(
            "p" * 140,
            "10001",
            MODULUS,
            "abcdef1234567890abcdef1234567890",
        )

        self.assertEqual(
            encrypted,
            "31ae2117e5a76245366fe7c9a650f5220ef815d0f89041710869e82e60e82b30"
            "902a4ab5cd3ccb89da54f5f899fa0c70b8fb5a627fc889a59cb260dbed66886d"
            "c3f03cce462b13658b5ca6f792dbc6f9cb7163aa530de67db8a89f36481ecf69"
            "d50ec84afbb3c165f302aa09e27c1f7cb542b84d4ddbcfe635fbd9b084764522 "
            "4323adffbeecbb646fa6e5898a140fce0384356078521f6bc2080ab4f95a96b8"
            "d6c45a32aeeb28e546ccf80bf2971b5cf7552e4a35fcc7e4a25a44090cc0a0ca"
            "27d95e4453f2ddb3fc67fecac545255a96fd7e32d704fe930277f89de22b2dcb"
            "3db28988a9fb684324e6a9f9a0c2efbb01a19c532d72f73687e902f9c4d7b94a",
        )
